/**
 * @file Defines the bottle in which the poor insects are stuck in.
 */

/**
 * All elements (position, width...) describe the inside of the bottle.
 * @typedef Bottle
 */

/**
 * @typedef Rectangle
 * @property {float} x
 * @property {float} y
 * @property {float} width
 * @property {float} height
 */

/** 
 * @param {Vector2D} position
 * @param {Rectangle} bounds  External bounds.
 **/
function BR_Bottle_create(position, width, height, bounds) {
	let bottle = {
		/** @type Vector2D **/
		position : position,

		width : width,
		height : height,

		// Wall in pixel
		thickness : BR_BOTTLE_WALL_THICKNESS,

		/** @type {ItemState} */
		dashState : BR_ItemState_create(),

		/** @type {Vector2D} */
		dashOrientation : { x : 0, y : 0 },

		/** @type {Rectangle} */
		bounds : bounds,

		graphicsZone : BR_Bottle_createBottleGraphics(position, width, height),
	};
	return bottle;
}

/** @return The graphics */
function BR_Bottle_createBottleGraphics(position, width, height) {
	let insideColor = BR_BOTTLE_COLOR_FLOOR;
	let wallColor = BR_BOTTLE_COLOR_WALL;
	let wallThickness = BR_BOTTLE_WALL_THICKNESS;

	// The zone in which the insects move.
	let insideWidth = width;
	let insideHeight = width;

	// The complete bottle.
	let outsideWidth = insideWidth + 2 * wallThickness;
	let outsideHeight = insideHeight + 2 * wallThickness;

	let borderThickness = BR_BOTTLE_BORDER_THICKNESS;
	let borderColor = BR_BOTTLE_BORDER_COLOR;

	let graphics = new PIXI.Graphics();

	// Draw the external border of the walls.
	{
		let x = 0;
		let y = 0;
		let width = outsideWidth;
		let height = outsideHeight;
		graphics.beginFill(borderColor);
		graphics.drawRect(x, y, width, height);
		graphics.endFill();
	}

	// Draw the walls.
	{
		let x = borderThickness;
		let y = borderThickness;
		let width = outsideWidth - 2 * borderThickness;
		let height = outsideHeight - 2 * borderThickness;
		graphics.beginFill(wallColor);
		graphics.drawRect(x, y, width, height);
		graphics.endFill();
	}

	// Draw the internal border of the wall.
	{
		let x = wallThickness - borderThickness;
		let y = wallThickness - borderThickness;
		let width = outsideWidth - 2 * (wallThickness - borderThickness);
		let height = outsideHeight - 2 * (wallThickness - borderThickness);
		graphics.beginFill(BR_BOTTLE_BORDER_COLOR);
		graphics.drawRect(x, y, width, height);
		graphics.endFill();
	}

	// Draw the inside.
	{
		let x = wallThickness;
		let y = wallThickness;
		let width = insideWidth;
		let height = insideHeight;
		graphics.beginFill(insideColor);
		graphics.drawRect(x, y, width, height);
		graphics.endFill();
	}

	graphics.position.x = position.x - wallThickness;
	graphics.position.y = position.y - wallThickness;

	return graphics;
}

/**
 * Start the dash. It should move the next frame.
 */
function BR_Bottle_startDash(bottle, left, right, up, down, level) {
	let orientation = { x : 0, y : 0 };

	if(left != right) {
		if(left) {
			orientation.x -= 1;
		} else {
			orientation.x += 1;
		}
	}

	if(up != down) {
		if(up) {
			orientation.y -= 1;
		} else {
			orientation.y += 1;
		}
	}

	let durationLevel = level;
	if(durationLevel >= BR_BOTTLE_DASH_LEVEL_DURATIONS.length) {
		level = BR_BOTTLE_DASH_LEVEL_DURATIONS.length - 1;
	}
	bottle.dashOrientation = BR_normalize(orientation);
	BR_ItemState_start(bottle.dashState, BR_BOTTLE_DASH_LEVEL_DURATIONS[level]);

	// TODO: cancel dash if already touching the bounds and going to them?
}

function BR_Bottle_isDashing(bottle) {
	return bottle.dashState.doing;
}

function BR_bottle_updateGraphicsZone(bottle) {
	bottle.graphicsZone.position.x = bottle.position.x - bottle.thickness;
	bottle.graphicsZone.position.y = bottle.position.y - bottle.thickness;
}

function BR_Bottle_dash(bottle, elapsedTime) {
	// Move.
	let orientation = bottle.dashOrientation
	let speedBySeconds = BR_BOTTLE_DASH_SPEED;
	let speed = speedBySeconds * elapsedTime;
	let deltaX = orientation.x * speed;
	let deltaY = orientation.y * speed;
	bottle.position.x += deltaX;
	bottle.position.y += deltaY;
	BR_ItemState_consume(bottle.dashState, elapsedTime);

	// Stay inside the bounds.
	{
		// Define the bottle outside values.
		let left = bottle.position.x - bottle.thickness;
		let right = bottle.position.x + bottle.width + bottle.thickness;
		let up = bottle.position.y - bottle.thickness;
		let down = bottle.position.y + bottle.height + bottle.thickness;

		let boundsLeft = bottle.bounds.x;
		let boundsRight = bottle.bounds.x + bottle.bounds.width;
		let boundsUp = bottle.bounds.y;
		let boundsDown = bottle.bounds.y + bottle.bounds.height;

		let margin = BR_MARGIN_BOTTLE;
		if(left <= boundsLeft) {
			bottle.position.x = boundsLeft + bottle.thickness + margin;
			BR_ItemState_reset(bottle.dashState);
		}
		if(boundsRight <= right) {
			bottle.position.x = boundsRight - bottle.width - bottle.thickness - margin;
			BR_ItemState_reset(bottle.dashState);
		}
		if(up <= boundsUp) {
			bottle.position.y = boundsUp + bottle.thickness + margin;
			BR_ItemState_reset(bottle.dashState);
		}
		if(boundsDown <= down) {
			bottle.position.y = boundsDown - bottle.height - bottle.thickness - margin;
			BR_ItemState_reset(bottle.dashState);
		}
	}
}
