/**
 * @file An insect. A beautiful and wonderful insect.
 */

/**
 * Defines the insect const stats.
 * @typedef InsectStats
 * @property {float} forwardSpeed
 * @property {float} backwardSpeed
 * @property {float} forwardSpeedWhileCharging
 * @property {float} backwardSpeedWhileCharging
 * @property {float} dashSpeed
 * @property {float} angleSpeed  In degrees.
 * @property {float} hitboxRadius2
 */

/**
 * @typedef Insect.
 */

/**
 * @param {InsectStats} stats
 */
function BR_Insect_create(stats, sprite, dashChargingLevelDuration, attackChargingLevelDuration, life) {
	let insect = {
		/** @type InsectStats */
		stats : stats,

		sprite : sprite,

		/** @type ChargingItem */
		dashCharging : BR_ChargingItem_create(dashChargingLevelDuration),

		/** @type {ItemState} */
		dashState : BR_ItemState_create(),

		/** @type {ChargingItem} */
		attackCharging : BR_ChargingItem_create(attackChargingLevelDuration),

		/** @type {ItemState} */
		attackState : BR_ItemState_create(),

		// Using circle hitboxes, all values are squared.
		hitboxRadius2 : stats.hitboxRadius2,

		life : life,
	};

	return insect;
}

function BR_Insect_reset(insect, dashChargingLevelDuration, attackChargingLevelDuration, life) {
	insect.sprite.angle = 0;
	insect.dashCharging = BR_ChargingItem_create(dashChargingLevelDuration);
	BR_ItemState_reset(insect.dashState);
	insect.attackCharging = BR_ChargingItem_create(attackChargingLevelDuration);
	BR_ItemState_reset(insect.attackState);
	insect.life = life;
	console.log("BR_Insect_reset life: " + life);
}

function BR_Insect_isDashing(insect) {
	return insect.dashState.doing;
}

function BR_Insect_isChargingDash(insect) {
	return insect.dashCharging.isCharging;
}

function BR_Insect_stopDash(insect) {
	BR_ItemState_reset(insect.dashState);
}

function BR_Insect_isCharging(insect) {
	return BR_Insect_isChargingDash(insect) || insect.attackCharging.isCharging;
}

function BR_Insect_dash(insect, elapsedTime) {
	let orientation = BR_getNormalOrientation(insect.sprite);
	let speedBySeconds = insect.stats.dashSpeed;
	let speed = speedBySeconds * elapsedTime;
	let deltaX = orientation.x * speed;
	let deltaY = orientation.y * speed;
	insect.sprite.x += deltaX;
	insect.sprite.y += deltaY;
	BR_ItemState_consume(insect.dashState, elapsedTime);
}

function BR_Insect_startDash(insect) {
	let level = insect.dashCharging.level;
	if(level > 0) {
		let remainingTime = BR_DASH_LEVEL_DURATIONS[level];
		BR_ItemState_start(insect.dashState, remainingTime)
	}
	BR_ChargingItem_reset(insect.dashCharging);
}

/**
 * @param {float} angleLimit  The angle limit to rotate (negative or positive).
 * @return {float} The actual rotation done.
 */
function BR_Insect_rotate(insect, angleLimit, elapsedTime) {
	if(angleLimit == 0) {
		return 0;
	}

	let angleToRotate = insect.stats.angleSpeed * elapsedTime;
	if(angleLimit > 0) {
		if(angleToRotate > angleLimit) {
			angleToRotate = angleLimit;
		}
	} else {
		angleToRotate = -angleToRotate;
		if(angleToRotate < angleLimit) {
			angleToRotate = angleLimit;
		}
	}

	insect.sprite.angle += angleToRotate;
	return angleToRotate;
}

function BR_Insect_moveForward(insect, elapsedTime) {
	let speedBySeconds = 0;
	if(BR_Insect_isCharging(insect)) {
		speedBySeconds = insect.stats.forwardSpeedWhileCharging;
	} else {
		speedBySeconds = insect.stats.forwardSpeed;
	}
	let speed = speedBySeconds * elapsedTime;
	let orientation = BR_getNormalOrientation(insect.sprite);
	let deltaX = orientation.x * speed;
	let deltaY = orientation.y * speed;
	insect.sprite.x += deltaX;
	insect.sprite.y += deltaY;
}

function BR_Insect_moveBackward(insect, elapsedTime) {
	let speedBySeconds = 0;
	if(BR_Insect_isCharging(insect)) {
		speedBySeconds = insect.stats.backwardSpeedWhileCharging;
	} else {
		speedBySeconds = insect.stats.backwardSpeed;
	}
	let speed = speedBySeconds * elapsedTime;
	let orientation = BR_getNormalOrientation(insect.sprite);
	let deltaX = orientation.x * speed;
	let deltaY = orientation.y * speed;
	insect.sprite.x -= deltaX;
	insect.sprite.y -= deltaY;
}

function BR_Insect_dead(insect) {
	// There is a bug: for some reason life is set to undefined.
	// This bug will be fixed... as soon as I use a real statically typed language!
	return insect.life <= 0 || insect.life == Math.NaN || insect.life == undefined;
}
function BR_Insect_getHit(insect, damage) {
	if(BR_Insect_dead(insect)) {
		return;
	}

	if(insect.life < damage) {
		insect.life = 0;
	} else {
		insect.life -= damage;
	}
}

/**
 * Is the insect going outside of the bottle, and if yes: where?
 * @return {object} Where it outbounds, if outbounds.
 */
function BR_Insect_outboundsBottle(insect, bottle) {
	let result = {
		left : false,
		right : false,
		up : false,
		down : false,
	};

	// Left.
	{
		let bottleLeft = bottle.position.x;
		let insectX = insect.sprite.position.x;
		let distance2 = BR_getDistance2({ x : insectX, y : 0 }, { x : bottleLeft, y : 0 });
		let crosses = distance2 <= insect.hitboxRadius2;
		let isOutside = insectX <= bottleLeft;
		result.left = (crosses || isOutside);
	}

	// Right.
	{
		let bottleRight = bottle.position.x + bottle.width;
		let insectX = insect.sprite.position.x;
		let distance2 = BR_getDistance2({ x : insectX, y : 0 }, { x : bottleRight, y : 0 });
		let crosses = distance2 <= insect.hitboxRadius2;
		let isOutside = bottleRight <= insectX;
		result.right = (crosses || isOutside);
	}

	// Up.
	{
		let bottleUp = bottle.position.y;
		let insectY = insect.sprite.position.y;
		let distance2 = BR_getDistance2({ x : 0, y : insectY }, { x : 0, y : bottleUp });
		let crosses = distance2 <= insect.hitboxRadius2;
		let isOutside = insectY <= bottleUp;
		result.up = (crosses || isOutside);
	}

	// Down.
	{
		let bottleDown = bottle.position.y + bottle.height;
		let insectY = insect.sprite.position.y;
		let distance2 = BR_getDistance2({ x : 0, y : insectY }, { x : 0, y : bottleDown });
		let crosses = distance2 <= insect.hitboxRadius2;
		let isOutside = bottleDown <= insectY;
		result.down = (crosses || isOutside);
	}

	return result;
}

/**
 * @return {BR_outboundsBottleResult}
 */
function BR_Insect_keepInBottle(insect, bottle) {
	const margin = BR_MARGIN_INSECT_WALL;

	let result = BR_Insect_outboundsBottle(insect, bottle);
	if(result.left) {
		let radius = Math.sqrt(insect.hitboxRadius2);
		let bottleLeft = bottle.position.x;
		insect.sprite.position.x = bottle.position.x + radius + margin;
	}
	if(result.right) {
		let radius = Math.sqrt(insect.hitboxRadius2);
		let bottleRight = bottle.position.x + bottle.width;
		insect.sprite.position.x = bottleRight - radius - margin;
	}
	if(result.up) {
		let radius = Math.sqrt(insect.hitboxRadius2);
		let bottleUp = bottle.position.y;
		insect.sprite.position.y = bottleUp + radius + margin;
	}
	if(result.down) {
		let radius = Math.sqrt(insect.hitboxRadius2);
		let bottleDown = bottle.position.y + bottle.height;
		insect.sprite.position.y = bottleDown - radius - margin;
	}

	return result;
}
